/* Copyright 2008 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <ctype.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/fd.h>

#include "virtio.h"
#include "kudzuint.h"
#include "modules.h"

static void virtioFreeDevice(struct virtioDevice *dev)
{
	freeDevice((struct device *) dev);
}

static void virtioWriteDevice(FILE *file, struct virtioDevice *dev)
{
	writeDevice(file, (struct device *)dev);
}

static int virtioCompareDevice(struct virtioDevice *dev1, struct virtioDevice *dev2)
{
	return compareDevice( (struct device *)dev1, (struct device *)dev2);
}

struct virtioDevice *virtioNewDevice(struct virtioDevice *old)
{
	struct virtioDevice *ret;

	ret = malloc(sizeof(struct virtioDevice));
	memset(ret, '\0', sizeof(struct virtioDevice));
	ret = (struct virtioDevice *) newDevice((struct device *) old, (struct device *) ret);
	ret->bus = BUS_VIRTIO;
	ret->newDevice = virtioNewDevice;
	ret->freeDevice = virtioFreeDevice;
	ret->writeDevice = virtioWriteDevice;
	ret->compareDevice = virtioCompareDevice;
	return ret;
}

static struct device *virtioProbeBlock(char *path, struct device *devlist)
{
	struct virtioDevice *vdev;

	vdev = virtioNewDevice(NULL);

	vdev->desc = strdup("Virtio Block Device");
	vdev->type = CLASS_HD;
	vdev->driver = strdup("virtio_blk");

	__getSysfsDevice((struct device *)vdev, path, "block:", 1);

	if (!vdev->device)
		vdev->device = strdup("vd");

	if (devlist)
		vdev->next = devlist;

	return (struct device *)vdev;
}

static struct device *virtioProbeNetwork(char *path, struct device *devlist)
{
	struct virtioDevice *vdev;

	vdev = virtioNewDevice(NULL);

	vdev->desc = strdup("Virtio Network Device");
	vdev->type = CLASS_NETWORK;
	vdev->driver = strdup("virtio_net");

	__getSysfsDevice((struct device *)vdev, path, "net:", 0);

	if (vdev->device)
		__getNetworkAddr((struct device *)vdev, vdev->device);
	else
		vdev->device = strdup("eth");

	if (devlist)
		vdev->next = devlist;

	return (struct device *)vdev;
}

static inline int getVirtioID(const char *dir)
{
	char path[256];

	snprintf(path, 256, "%s/%s", dir, "device");

	return __readInt(path);
}

struct device *virtioProbe(enum deviceClass probeClass, int probeFlags,
			   struct device *devlist)
{
	struct dirent *ent;
	DIR *dir;

	if (!(probeClass & CLASS_HD) && !(probeClass & CLASS_NETWORK))
		return devlist;

	if (access("/sys/bus/virtio/devices", R_OK))
		return devlist;

	dir = opendir("/sys/bus/virtio/devices");

	while ((ent = readdir(dir))) {
		char path[256];
		int id;

		snprintf(path, 256, "/sys/bus/virtio/devices/%s", ent->d_name);

		id = getVirtioID(path);

		switch (id) {
		case VIRTIO_ID_NET:
			if (probeClass & CLASS_NETWORK)
				devlist = virtioProbeNetwork(path, devlist);
			break;

		case VIRTIO_ID_BLOCK:
			if (probeClass & CLASS_HD)
				devlist = virtioProbeBlock(path, devlist);
			break;

		default:
			break;
		}
	}

	closedir(dir);

	return devlist;
}
